@@ -1,12 +1,14 @@
 use std::collections::HashMap;
 use std::collections::hash_map::Entry;
-use std::fs::File;
+use std::fs::{DirBuilder, OpenOptions};
 use std::sync::Arc;
 use std::sync::Weak;
 use std::ops::Deref;
 use std::ops::DerefMut;
 use std::path::Path;
 use std::path::PathBuf;
+#[cfg(unix)]
+use std::os::unix::fs::OpenOptionsExt;
 
 use futures::lock::Mutex;
 use futures::lock::MutexGuard;
@@ -236,6 +238,9 @@ impl Backend {
             })),
         };
 
+        // Restrict access to the home directory to the current user.
+        backend.restrict_access().await;
+
         if let Err(err) = backend.scan().await {
             log::debug!("Scanning {:?} for keys: {}", home, err);
         }
@@ -243,15 +248,7 @@ impl Backend {
         Ok(Box::new(backend))
     }
 
-    /// Initializes a SoftKeys backend.
-    ///
-    /// `home` is the directory where the backend will look for its
-    /// configuration.  If this is `None`, then it will look in
-    /// `$HOME/.sq/keystore/softkeys`.
-    ///
-    /// Currently, this backend looks for certificates containing
-    /// secret key material in files with an 'asc' or 'pgp' extension.
-    /// Other files are silently ignored.
+    /// Initializes an ephemeral SoftKeys backend.
     pub async fn init_ephemeral() -> Result<Self> {
         log::trace!("Backend::init_ephemeral");
 
@@ -267,12 +264,38 @@ impl Backend {
             })),
         };
 
+        // Restrict access to the home directory to the current user.
+        backend.restrict_access().await;
+
         if let Err(err) = backend.scan().await {
             log::error!("Scanning {:?} for keys: {}", home, err);
         }
 
         Ok(backend)
     }
+
+
+    /// Restricts access to the home directory to the current user.
+    ///
+    /// This is a best-effort mechanism.  And, the home directory
+    /// might not even exist at this point.
+    async fn restrict_access(&self) {
+        fn restrict(home: &Path) -> Result<()> {
+            #[cfg(unix)]
+            {
+                use std::fs;
+                use std::os::unix::fs::PermissionsExt;
+
+                let mut perms = fs::metadata(home)?.permissions();
+                perms.set_mode(0o700);
+                fs::set_permissions(home, perms)?;
+            }
+
+            Ok(())
+        }
+
+        let _ = restrict(&self.lock().await.home);
+    }
 }
 
 impl BackendLockGuard<'_> {
@@ -541,8 +564,29 @@ impl backend::Backend for Backend {
 
         let mut backend = self.lock().await;
 
-        std::fs::create_dir_all(&backend.home)
-            .with_context(|| format!("creating {}", backend.home.display()))?;
+        // Create all directories up to but excluding the last
+        // component.
+        if let Some(path) = backend.home.parent() {
+            DirBuilder::new()
+                .recursive(true)
+                .create(path)
+                .with_context(
+                    || format!("creating {}", path.display()))?;
+        }
+
+        // And now create the last component.
+        if ! backend.home.exists() {
+            let mut builder = DirBuilder::new();
+            #[cfg(unix)]
+            {
+                // Restrict access to the current user.
+                use std::os::unix::fs::DirBuilderExt;
+                builder.mode(0o700);
+            }
+            builder.create(&backend.home)
+                .with_context(
+                    || format!("creating {}", backend.home.display()))?;
+        }
 
         let fpr = cert.fingerprint();
 
@@ -592,13 +636,10 @@ impl backend::Backend for Backend {
             // place.  This doesn't eliminate races, but it does
             // prevent a partial update from destroying the existing
             // data.
-            let mut tmp_filename = filename.clone();
-            tmp_filename.set_extension("pgp~");
-
-            let mut f = File::create(&tmp_filename)?;
-            cert.as_tsk().serialize(&mut f)?;
-
-            std::fs::rename(&tmp_filename, &filename)?;
+            let mut f = RestrictedNamedTempFile::new_in(
+                filename.parent().expect("file in softkeys"))?;
+            cert.as_tsk().serialize(f.file())?;
+            f.persist(&filename)?;
         }
 
         backend.ingest(cert, filename, true).await
@@ -1023,23 +1064,24 @@ impl backend::KeyHandle for Key {
                 })?.0
         };
 
-        let mut tmp_filename = variant.path.clone();
-        tmp_filename.set_extension("pgp~");
-        log::debug!("Writing updated certificate to {}",
-                    tmp_filename.display());
-
-        let mut f = File::create(&tmp_filename)
+        let mut f = RestrictedNamedTempFile::new_in(
+            variant.path.parent().expect("file in softkeys"))
             .with_context(|| {
-                format!("Creating {}", tmp_filename.display())
+                format!("Creating temporary file in {}",
+                        variant.path.parent().expect("file in softkeys")
+                        .display())
             })?;
-        cert.as_tsk().serialize(&mut f)
+
+        log::debug!("Writing updated certificate to {}",
+                    f.path().display());
+        cert.as_tsk().serialize(f.file())
             .with_context(|| {
-                format!("Updating {}", tmp_filename.display())
+                format!("Updating {}", f.path().display())
             })?;
 
         log::debug!("Moving {} to {}",
-                    tmp_filename.display(), variant.path.display());
-        std::fs::rename(&tmp_filename, &variant.path)
+                    f.path().display(), variant.path.display());
+        f.persist(&variant.path)
             .with_context(|| {
                 format!("Updating {}", variant.path.display())
             })?;
@@ -1114,23 +1156,23 @@ impl backend::KeyHandle for Key {
                 log::debug!("Removing {} from {}.",
                             fpr, variant.path.display());
 
-                let mut tmp_filename = variant.path.clone();
-                tmp_filename.set_extension("pgp~");
-                log::debug!("Writing updated certificate to {}",
-                            tmp_filename.display());
-
-                let mut f = File::create(&tmp_filename)
+                let mut f = RestrictedNamedTempFile::new_in(
+                    variant.path.parent().expect("file in softkeys"))
                     .with_context(|| {
-                        format!("Creating {}", tmp_filename.display())
+                        format!("Creating temporary file in {}",
+                                variant.path.parent().expect("file in softkeys")
+                                .display())
                     })?;
-                cert.as_tsk().serialize(&mut f)
+                log::debug!("Writing updated certificate to {}",
+                            f.path().display());
+                cert.as_tsk().serialize(f.file())
                     .with_context(|| {
-                        format!("Updating {}", tmp_filename.display())
+                        format!("Updating {}", f.path().display())
                     })?;
 
                 log::debug!("Moving {} to {}",
-                            tmp_filename.display(), variant.path.display());
-                std::fs::rename(&tmp_filename, &variant.path)
+                            f.path().display(), variant.path.display());
+                f.persist(&variant.path)
                     .with_context(|| {
                         format!("Updating {}", variant.path.display())
                     })?;
@@ -1178,6 +1220,66 @@ impl backend::KeyHandle for Key {
     }
 }
 
+/// Similar to `tempfile::NamedTempFile`, but restricts access to the
+/// created files to the current user on UNIX.
+struct RestrictedNamedTempFile {
+    path: PathBuf,
+    file: std::fs::File,
+}
+
+impl RestrictedNamedTempFile {
+    /// Creates a new temporary file in the given directory.
+    fn new_in(dir: &Path) -> Result<Self> {
+        loop {
+            let mut r = [0; 4];
+            openpgp::crypto::random(&mut r)?;
+            let name =
+                format!("{}.tmp", openpgp::fmt::hex::encode(&r));
+            let path = dir.join(name);
+
+            let mut o = OpenOptions::new();
+            o.create_new(true)
+                .write(true);
+            #[cfg(unix)]
+            o.mode(0o600);
+
+            match o.open(&path) {
+                Ok(file) => return Ok(RestrictedNamedTempFile {
+                    path,
+                    file,
+                }),
+
+                Err(e) if e.kind() == std::io::ErrorKind::AlreadyExists =>
+                    continue, // Retry.
+
+                Err(e) => return Err(e.into()),
+            }
+        }
+    }
+
+    /// Returns the temporary path.
+    fn path(&self) -> &Path {
+        &self.path
+    }
+
+    /// Returns a mutable reference to the open file.
+    fn file(&mut self) -> &mut std::fs::File {
+        &mut self.file
+    }
+
+    /// Persists the temporary file under the given name.
+    fn persist(self, path: &Path) -> Result<()> {
+        std::fs::rename(&self.path, path)?;
+        Ok(())
+    }
+}
+
+impl Drop for RestrictedNamedTempFile {
+    fn drop(&mut self) {
+        let _ = std::fs::remove_file(&self.path);
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use openpgp::KeyHandle;
